package be.recipe.api.patient;

import static be.recipe.api.constraints.Constraint.required;
import static be.recipe.api.prescriber.VisionOtherPrescribers.OPEN;
import static be.recipe.api.reservation.Reserved.no;
import static be.recipe.api.reservation.Reserved.yes;
import static be.recipe.api.viewer.AllViewer.all;
import static java.util.Arrays.asList;

import be.recipe.api.Prescription;
import be.recipe.api.Registered;
import be.recipe.api.executor.Executor;
import be.recipe.api.executor.ExecutorViewer;
import be.recipe.api.prescriber.VisionOtherPrescribers;
import be.recipe.api.reservation.ContactPreference;
import be.recipe.api.reservation.Reservation;
import be.recipe.api.reservation.Reservation.Status;
import be.recipe.api.reservation.Reserved;
import be.recipe.api.series.Page;
import be.recipe.api.series.PartialResult;
import be.recipe.api.series.SortedBy;
import java.util.Collections;
import java8.util.Optional;
import java8.util.function.Consumer;
import java8.util.function.Function;
import java8.util.stream.Collectors;
import java8.util.stream.Stream;
import java8.util.stream.StreamSupport;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import org.threeten.bp.LocalDateTime;

// tag::class[]
public interface ListPrescriptions<R extends ListPrescriptions.Request> {
  // end::class[]
  // tag::methods[]
  PartialResult<Response> list(R request);
  // end::methods[]
  // tag::class[]

  class Request {
    // end::class[]
    // tag::attributes[]
    public Reserved reserved;

    @Valid
    @NotNull(message = required)
    public Page page = Page.first();

    public SortedBy<Prescription.Attribute>[] sortedBy;
    // end::attributes[]

    public static Request request() {
      return new Request();
    }

    @SafeVarargs
    public static Request sortedBy(SortedBy<Prescription.Attribute>... sortedBy) {
      final Request request = new Request();
      StreamSupport.stream(asList(sortedBy))
          .forEach(
              new Consumer<SortedBy<Prescription.Attribute>>() {
                @Override
                public void accept(SortedBy<Prescription.Attribute> it) {
                  request.sortedBy(it);
                }
              });
      return request;
    }

    public Request() {}

    public Request(Page page) {
      this.page = page;
    }

    public Request reserved(Reserved reserved) {
      this.reserved = reserved;
      return this;
    }

    public Request reserved() {
      return reserved(yes);
    }

    public Request notReserved() {
      return reserved(no);
    }

    @SuppressWarnings({"unchecked", "UnusedReturnValue"})
    public Request sortedBy(SortedBy<Prescription.Attribute> sortedBy) {
      this.sortedBy =
          StreamSupport.stream(
                  asList(
                      StreamSupport.stream(Collections.singleton(sortedBy)),
                      this.sortedBy == null
                          ? StreamSupport.stream(
                              Collections.<SortedBy<Prescription.Attribute>>emptyList())
                          : StreamSupport.stream(asList(this.sortedBy))))
              .flatMap(
                  new Function<
                      Stream<SortedBy<Prescription.Attribute>>,
                      Stream<SortedBy<Prescription.Attribute>>>() {
                    @Override
                    public Stream<SortedBy<Prescription.Attribute>> apply(
                        Stream<SortedBy<Prescription.Attribute>> stream) {
                      return stream;
                    }
                  })
              .collect(Collectors.<SortedBy<Prescription.Attribute>>toList())
              .toArray(new SortedBy[0]);
      return this;
    }
    // tag::class[]
  }

  interface Response extends Registered {
    // end::class[]
    // tag::methods[]
    ExecutorViewer visibleToExecutors();

    VisionOtherPrescribers visionOtherPrescribers();

    Prescription.Status status();

    LocalDateTime creationDate();

    Optional<Reservation> reservation();
    // end::methods[]
    // tag::class[]

    interface Reservation {
      // end::class[]
      // tag::methods[]
      Executor.ID executorId();

      Status status();

      LocalDateTime creationDateTime();

      boolean hasUpdates();

      String feedbackToPatient();

      String email();

      String phoneNumber();

      ContactPreference contactPreference();
      // end::methods[]

      class Simple implements Reservation {
        public Executor.ID executorId;
        public LocalDateTime creationDateTime;
        public ContactPreference contactPreference;
        public String feedbackToPatient, email, phoneNumber;
        public boolean hasUpdates;
        public Status status;

        @Override
        public Executor.ID executorId() {
          return executorId;
        }

        @Override
        public Status status() {
          return status;
        }

        @Override
        public LocalDateTime creationDateTime() {
          return creationDateTime;
        }

        @Override
        public boolean hasUpdates() {
          return hasUpdates;
        }

        @Override
        public String feedbackToPatient() {
          return feedbackToPatient;
        }

        @Override
        public String email() {
          return email;
        }

        @Override
        public String phoneNumber() {
          return phoneNumber;
        }

        @Override
        public ContactPreference contactPreference() {
          return contactPreference;
        }
      }
      // tag::class[]
    }
    // end::class[]

    class Simple implements Response {
      public be.recipe.api.patient.ListPrescriptions.Response.Reservation reservation;
      public VisionOtherPrescribers visionOtherPrescribers = OPEN;
      public ExecutorViewer visibleToExecutors = all();
      public Prescription.Status status;
      public LocalDateTime creationDate;
      public Prescription.RID rid;

      public static Response response(Prescription.Encrypted target) {
        return new PrescriptionWrapper(target);
      }

      public static Response.Simple response(Prescription.RID rid) {
        Response.Simple it = new Response.Simple();
        it.rid = rid;
        return it;
      }

      @Override
      public Prescription.RID rid() {
        return rid;
      }

      @Override
      public ExecutorViewer visibleToExecutors() {
        return visibleToExecutors;
      }

      @Override
      public VisionOtherPrescribers visionOtherPrescribers() {
        return visionOtherPrescribers;
      }

      @Override
      public Prescription.Status status() {
        return status;
      }

      @Override
      public LocalDateTime creationDate() {
        return creationDate;
      }

      @Override
      public Optional<Reservation> reservation() {
        return Optional.ofNullable(reservation);
      }
    }

    class Wrapper implements Response {
      private final Response target;

      public Wrapper(Response target) {
        this.target = target;
      }

      @Override
      public Prescription.RID rid() {
        return target.rid();
      }

      @Override
      public ExecutorViewer visibleToExecutors() {
        return target.visibleToExecutors();
      }

      @Override
      public VisionOtherPrescribers visionOtherPrescribers() {
        return target.visionOtherPrescribers();
      }

      @Override
      public Prescription.Status status() {
        return target.status();
      }

      @Override
      public LocalDateTime creationDate() {
        return target.creationDate();
      }

      @Override
      public Optional<Reservation> reservation() {
        return target.reservation();
      }
    }
    // tag::class[]
  }
}
// end::class[]

class PrescriptionWrapper implements ListPrescriptions.Response {
  private final Prescription.Encrypted target;

  public PrescriptionWrapper(Prescription.Encrypted target) {
    this.target = target;
  }

  @Override
  public Prescription.RID rid() {
    return target.rid();
  }

  @Override
  public ExecutorViewer visibleToExecutors() {
    return target.visibleToExecutors();
  }

  @Override
  public VisionOtherPrescribers visionOtherPrescribers() {
    return target.visionOtherPrescribers();
  }

  @Override
  public Prescription.Status status() {
    return target.status();
  }

  @Override
  public LocalDateTime creationDate() {
    return target.creationDate();
  }

  @Override
  public Optional<ListPrescriptions.Response.Reservation> reservation() {
    return target
        .reservation()
        .map(
            new Function<be.recipe.api.reservation.Reservation, Reservation>() {
              @Override
              public Reservation apply(be.recipe.api.reservation.Reservation reservation) {
                return new ReservationWrapper(reservation);
              }
            });
  }
}

class ReservationWrapper implements ListPrescriptions.Response.Reservation {
  private final Reservation target;

  ReservationWrapper(Reservation target) {
    this.target = target;
  }

  @Override
  public Executor.ID executorId() {
    return target.executor();
  }

  @Override
  public Status status() {
    return target.status();
  }

  @Override
  public LocalDateTime creationDateTime() {
    return target.creationDateTime();
  }

  @Override
  public boolean hasUpdates() {
    return target.newInfoForPatient();
  }

  @Override
  public String feedbackToPatient() {
    return target.feedbackToPatient();
  }

  @Override
  public String email() {
    return target.email();
  }

  @Override
  public String phoneNumber() {
    return target.phoneNumber();
  }

  @Override
  public ContactPreference contactPreference() {
    return target.contactPreference();
  }
}
